import sys
import os
import time
import logging
import tempfile
import subprocess
import _thread
import traceback

import tornado.web
import tornado.ioloop
import tornado.gen
import requests
import psutil

from .ssh import SSHRemoteForward

log = logging.getLogger('curlbomb.server')

class CurlbombBaseRequestHandler(tornado.web.RequestHandler):
    """Base RequestHandler

      Implementations:
       - CurlbombResourceRequestHandler 
       - CurlbombStreamRequestHandler
    """
    
    def initialize(self, resource, state, allowed_gets=1, knock=None,
                   mime_type='text/plain', allow_post_backs=False,
                   log_post_backs=False, log_file=None, get_callback=None):
        """Arguments:
        
          resource         - A file like object to serve the contents of
          state            - State dictionary to maintain across requests
          allowed_gets     - Number of gets allowed before quiting
          knock            - The required X-knock header the client must send, or None
          mime_type        - The mime type the server should declare the content as
          allow_post_backs - Allow client to post data back to the server. 
                             Delays server termination until post_backs == allowed_gets
          log_post_backs   - Log post backs to stdout
          log_file         - Log post backs to file
          get_callback     - callback to run when get is finished, passes request
          *args            - The rest of the RequestHandler args
          **kwargs         - The rest of the RequestHandler kwargs
        """
        self._resource = resource
        self._allowed_gets = allowed_gets
        self._knock = knock
        self._mime_type = mime_type
        self._allow_post_backs = allow_post_backs
        self._get_callback = get_callback
        self._log_post_backs = log_post_backs
        self._log_file = log_file
        
        self._state = state
                
    def prepare(self):
        self.request.start_time = time.time()
        # Validate X-knock header if one is required:
        if self._knock is not None:
            x_knock = self.get_argument(
                'knock', self.request.headers.get('X-knock', None))
            if x_knock != self._knock:
                log.info("Invalid knock")
                raise tornado.web.HTTPError(401, 'Invalid knock')
    
    def shutdown_if_ready(self):
        """Shutdown if it's time to shutdown"""
        # If the resource has been retrieved all the times it's allowed:
        if self._allowed_gets > 0 and self._allowed_gets <= self._state['num_gets']:
            num_post_backs = self._state['num_posts']
            # If we're still waiting for post backs:
            if self._allow_post_backs and num_post_backs < self._allowed_gets:
                log.info("Waiting for {} more post backs from client".format(
                    self._allowed_gets - num_post_backs))
            else:
                # Shutdown:
                log.info("Served resource {} times. Done. Waiting for network buffers to clear".format(self._state['num_gets']))
                # Query psutil for ongoing socket connections that we don't want to kill yet:
                proc = psutil.Process(os.getpid())
                while True:
                    for conn in proc.connections():
                        if conn.status == 'ESTABLISHED':
                            time.sleep(1)
                            break
                    else:
                        break    
                tornado.ioloop.IOLoop.current().stop()

    def write_error(self, status_code, **kwargs):
        try:
            log_message = kwargs.get('exc_info')[1].log_message
        except (TypeError, AttributeError, IndexError):
            log_message = 'unknown reason'
        self.finish(log_message+'\r\n')

class CurlbombResourceWrapperRequestHandler(tornado.web.RequestHandler):
    """Serve a script that wraps another curlbomb"""
    def initialize(self, curlbomb_command):
        self.__curlbomb_command = curlbomb_command
        
    def get(self):
        self.set_status(200)
        self.write(self.__curlbomb_command)
        self.finish()

    def head(self):
        """Allow head requests, does not count towards num_gets and does not require a knock"""
        pass
                
class CurlbombResourceRequestHandler(CurlbombBaseRequestHandler):
    """Serve a file like resource a limited number of times.
    Allow response data to be posted back."""
    def get(self):
        if self._allowed_gets == 0 or self._state['num_gets'] < self._allowed_gets:
            # Client is allowed to get:
            self.set_status(200)
            self.add_header("Content-type", self._mime_type)
            self.write(self._resource.read())
            self._state['num_gets'] += 1
        else:
            # Client is not allowed to get any more:
            log.info("Resource denied (max gets reached) to: {}".format(
                self.request.remote_ip))
            raise tornado.web.HTTPError(405, 'Client is not allowed to GET anymore')
        self.finish()
        if self._state['num_gets'] < self._allowed_gets or self._allowed_gets == 0:
            self._resource.seek(0)
        if self._get_callback is not None:
            self._get_callback(self.request)
        self.shutdown_if_ready()

    def head(self):
        """Allow head requests, does not count towards num_gets and does not require a knock"""
        pass


@tornado.web.stream_request_body
class CurlbombStreamRequestHandler(CurlbombBaseRequestHandler):
    """Stream output of script from client back to the server"""    
    def data_received(self, data):
        """Handle incoming PUT data"""
        if self._log_post_backs:
            sys.stdout.buffer.write(data)
        if self._log_file:
            self._log_file.write(data)
            self._log_file.flush()

    def put(self):
        """Finish streamed PUT request"""
        self._state['num_posts_in_progress'] -= 1
        self._state['num_posts'] += 1
        self.finish()
        self.shutdown_if_ready()

    def post(self):
        self.put()
        
    def prepare(self):
        CurlbombBaseRequestHandler.prepare(self)
        log.info("Stream prepare")
        if not self._allow_post_backs:
            raise tornado.web.HTTPError(405, 'This server is not configured to allow data upload')
        if (self._state['num_posts'] +
            self._state['num_posts_in_progress']) >= self._allowed_gets and self._allowed_gets != 0:
            raise tornado.web.HTTPError(403, 'Maximum number of posts reached')
        self._state['num_posts_in_progress'] += 1
                
class ErrorRequestHandler(tornado.web.RequestHandler):
    def get(self):
        raise tornado.web.HTTPError(404, 'Not Found')



def run_server(settings):
    settings['state'] = {'num_gets': 0, 'num_posts': 0, 'num_posts_in_progress': 0}
    curlbomb_args = dict(
        resource=settings['resource'],
        state=settings['state'],
        allowed_gets=settings['num_gets'],
        knock=settings['knock'],
        mime_type=settings['mime_type'],
        allow_post_backs=settings['receive_postbacks'],
        log_post_backs=settings['log_post_backs'],
        log_file=settings['log_file'],
        get_callback=settings.get('get_callback', None)
    )
    
    unwrapped_script = settings['get_curlbomb_command'](settings, unwrapped=True)
    if not settings['client_quiet'] and settings['time_command']:
        unwrapped_script = "time "+unwrapped_script

    app = tornado.web.Application(
        [
            (r"/", CurlbombResourceWrapperRequestHandler,
             dict(curlbomb_command=unwrapped_script)),
            (r"/r", CurlbombResourceRequestHandler, curlbomb_args),
            (r"/s", CurlbombStreamRequestHandler, curlbomb_args)
        ], default_handler_class=ErrorRequestHandler
    )

    global httpd
    httpd = app.listen(
        settings['port'],
        ssl_options=settings['ssl_context'],
        max_buffer_size=1024E9)

    ## Start SSH tunnel if requested:
    httpd.ssh_conn = None
    if settings['ssh']:
        if settings['ssl'] is False:
            log.warn("Using --ssh without --ssl is probably not a great idea")
        httpd.ssh_conn = SSHRemoteForward(
            settings['ssh_host'], settings['ssh_forward'], settings['ssh_port'])
        httpd.ssh_conn.start()
        if not httpd.ssh_conn.wait_connected():
            log.error(httpd.ssh_conn.last_msg)
            sys.exit(1)

    cmd = settings['get_curlbomb_command'](settings)
    if not settings['quiet']:
        if settings['stdout'].isatty():
            sys.stderr.write("Paste this command on the client:\n\n")
            sys.stderr.write("  {}\n\n".format(cmd))
            if settings['passphrase']:
                sys.stderr.write("Client passphrase: {}\n\n".format(settings['passphrase']))
        else:
            sys.stderr.write("{}\n".format(cmd))

    # Disable port forward checker for now. Good idea, but it doesn't work reliably.
    # 
    # if settings['ssh'] and not settings['args'].domain:
    #     "Check the SSH forward works"
    #     def check_port_forward(timeout=5):
    #         time.sleep(5)
    #         try:
    #             url = "http{ssl}://{host}:{port}".format(
    #                 ssl="s" if settings['ssl'] is not False else "",
    #                 host=settings['display_host'],
    #                 port=settings['display_port'])
    #             log.info("Testing port forward is functioning properly - {}".format(url))
    #             r = requests.head(url, timeout=timeout)
    #         except (requests.ConnectionError, requests.exceptions.ReadTimeout):
    #             log.warn("Could not contact server throuh SSH forward. You may need to check your sshd_config and enable 'GatwayPorts clientspecified'")
    #     _thread.start_new_thread(check_port_forward, ())

    try:
        log.debug("server ready on local port {}".format(settings['port']))
        tornado.ioloop.IOLoop.current().start()
    except KeyboardInterrupt:
        if settings['verbose']:
            traceback.print_exc()
    finally:
        httpd.stop()
        if httpd.ssh_conn is not None:
            httpd.ssh_conn.kill()
        settings['resource'].close()
        if settings['log_process']:
            settings['log_file'].close()
            settings['log_process'].wait()
            log.info("run_server done")

    return settings.get('return_code', 0)



import sys
import argparse
import os

from . import run
from . import get
from . import put
from . import ping
from . import ssh_copy_id

def get_version(include_path=False):
    import pkg_resources
    path = os.path.split(__file__)[0]
    try:
        pkg = pkg_resources.get_distribution('curlbomb')
    except pkg_resources.DistributionNotFound:
        return 'DEV' + ( " - {}".format(path) if include_path else "")
    if __file__.startswith(pkg.location):
        return pkg.version + ( " - {}".format(path) if include_path else "")
    return 'DEV' + ( " - {}".format(path) if include_path else "")

def argparser(formatter_class=argparse.HelpFormatter):
    parser = argparse.ArgumentParser(
        description='curlbomb is an HTTP server for serving one-time-use shell scripts',
        formatter_class=formatter_class)
    subparsers = parser.add_subparsers()
    
    parser.add_argument('-n', '--num-gets', metavar="N",
                        help="Number of times to serve resource (default:1)",
                        type=int, default=1)
    parser.add_argument('-p', '--port',  help="TCP port number to use "
                        "(default:random available)",
                        default="random")
    parser.add_argument('-d','--domain', metavar="host[:port]",
                        help="Provide the domain and port to display "
                        "in the constructed URL. (example.com:8080)")
    parser.add_argument('-w', '--wget', 
                        help="Output wget command rather than curl",
                        action="store_true")
    parser.add_argument('-l','--log-posts', dest="log_post_backs", action="store_true",
                        help="Log client stdout to server stdout")
    parser.add_argument('-q', '--quiet', action="store_true",
                        help="Be more quiet. Don't print the curlbomb command")
    parser.add_argument('-v', '--verbose', action="store_true",
                        help="Be more verbose. Enables --log-posts and print INFO logging")
    parser.add_argument('--ssh', metavar="SSH_FORWARD",
                        help="Forward curlbomb through another host via SSH - "
                        "[user@]host[:ssh_port][:http_port]",
                        default=None)
    parser.add_argument('--ssl', metavar="CERTIFICATE",
                        help="Use SSL with the given certificate file "
                        "(optionally PGP encrypted.) CERTIFICATE may be specified as a "
                        "single - to generate a new self-signed certificate and to turn "
                        "on --pin", default=False)
    parser.add_argument('--pin', help="Pin the SSL certificate hash into the client "
                        "command to force curl to use our certificate"
                        " (requires --ssl)", action="store_true")
    parser.add_argument("-e", "--encrypt", action="store_true",
                        help="Encrypt resource with GPG before serving. Will use a randomly generated "
                        "symmetric passphrase unless --encrypt-to or --passphrase is specified.")
    parser.add_argument("--encrypt-to", action="append", metavar="GPG_ID", help="Encrypt with the "
                        "public key specified instead of a passphrase. Can be specified multiple times.")
    parser.add_argument("--passphrase", action="store_true",
                        help="Ask for a symmetric passphrase to encrypt with instead of a random one.")
    parser.add_argument('--survey', help="Just a survey mission, no bomb run "
                        "(just get the script, don't run it)", action="store_true")
    parser.add_argument('--unwrapped',
                        help="Get the unwrapped version of the curlbomb "
                        "(1 less server request, but longer command)", action="store_true")
    parser.add_argument('--client-logging', dest="client_logging",
                        help="Enable client execution log (curlbomb.log on client)",
                        action="store_true")
    parser.add_argument('--client-quiet', dest="client_quiet",
                        help="Quiet the output on the client",
                        action="store_true")
    parser.add_argument('--mime-type', help="The content type to serve",
                        default="text/plain")
    parser.add_argument('--pipe', help="Pipe to shell command rather than doing process substitution. "
                        "This is necessary for most interactive scripts.",
                        action="store_true")
    parser.add_argument('--disable-knock', action="store_true",
                        help="Don't require authentication (no X-knock header)")
    parser.add_argument('--knock', help="Use a specific knock rather than random",
                        default=None)
    parser.add_argument('--debug', action="store_true",
                        # Be really verbose, turn on debug logging
                        help=argparse.SUPPRESS)
    # Old disabled method for disabling postback logging:
    # Now this just outputs an error if the user uses it:
    parser.add_argument('-1', '--disable-postback', dest='disable_postback',
                        action="store_true", help=argparse.SUPPRESS)
    parser.add_argument('--version', action="version", version=get_version(True))

    
    run.add_parser(subparsers)
    put.add_parser(subparsers)
    get.add_parser(subparsers)
    ping.add_parser(subparsers)
    ssh_copy_id.add_parser(subparsers)
    
    return parser





from io import BytesIO
from collections import OrderedDict
import urllib
import shlex
import subprocess
import logging

log = logging.getLogger('curlbomb.ping')

def add_parser(subparsers):
    ping_parser = subparsers.add_parser(
        'ping', help="Waits for client(s) to make a request, containing optional "
        "message and return parameters. Returns 0 or the last non-zero return "
        "parameter received from client(s).")

    def return_code(x):
        x = int(x)
        if x<0 or x>255:
            raise ValueError("valid range is 0-255")
        return x

    ping_parser.add_argument('-m', '--message',
                             help="Adds message parameter to ping request")
    ping_parser.add_argument('-r', '--return', dest='return_code',
                             type=return_code,
                             help="Adds return parameter to ping request")
    ping_parser.add_argument(
        '--return-success', action='store_true',
        help="Always return 0 regardless of the 'return' parameter the "
        "client(s) sends back")
    ping_parser.add_argument('-c','--command', help="Command to run on ping. "
                             "string formatters include: {return_code}, {message} "
                             "(don't use quotes around them)")
    ping_parser.add_argument('-n', '--notify', action="store_true",
                             help="Notify of ping via libnotify (python-notify2 package)")
    ping_parser.set_defaults(prepare_command=prepare)

def prepare(args, settings, parser):
    settings['resource'] = BytesIO(b'')
    settings['survey'] = True
    settings['receive_postbacks'] = False
    settings['server_shutdown_timeout'] = 0
    
    def get_ping_command(settings, unwrapped=None):
        params = OrderedDict()
        if settings['require_knock']:
            params['knock'] = settings['knock']
        if args.message:
            params['message'] = args.message
        if args.return_code:
            params['return'] = args.return_code
        
        return "curl -LSs 'http{ssl}://{host}:{port}/r{query_params}'".format(
            ssl="s" if settings['ssl'] is not False else "",
            host=settings['display_host'],
            port=settings['display_port'],
            query_params="?"+urllib.parse.urlencode(
                params) if len(params)>0 else ""
            )
    settings['get_curlbomb_command'] = get_ping_command
    
    def get_callback(request):
        """Callback that server runs on ping from client

        request - the tornado.web.HTTPRequest from the client
        """
        # Handle return code parameter:
        return_code = request.arguments.get('return', [b"0"])[0]
        message = request.arguments.get('message', [b""])[0].decode("utf-8")
        try:
            return_code = int(return_code)
        except ValueError:
            log.warn("Client ping specified non-integer return code: {}".format(
                return_code))
            return_code = 0
        # Only change the return code if it's not 0.
        # This way multiple clients can ping and
        # all must be successful to return 0:
        if return_code != 0 and not args.return_success:
            settings['return_code'] = return_code

        # Handle notification command (-c)
        if args.command is not None:
            command = args.command.format(
                return_code=return_code, message='"{}"'.format(message.replace(r'"',r'\"')))
            log.info("Running notification command: {}".format(shlex.split(command)))
            out = subprocess.Popen(shlex.split(command),
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT).communicate()[0]
            log.info("notification out: {}".format(out))

        # Handle desktop notification (-n)
        if args.notify:
            try:
                import notify2
                notify2.init("curlbomb")
            except ImportError:
                log.error("Desktop notifications are disabled. Please install "
                          "python-notify2 package to enable.")
            else:
                notify2.Notification("Ping", message).show()
                
    settings['get_callback'] = get_callback



import subprocess
import tempfile
import logging
import ssl
import base64

from . import gpg

log = logging.getLogger('curlbomb.ssl')

def decrypt_cert_if_necessary(cert):
    """GPG decrypt cert pem"""
    if gpg.detect_encrypted_resource(cert):
        log.info("Attempting SSL certificate decryption")
        with subprocess.Popen(
                ['gpg','-d'], stdin=subprocess.PIPE,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE) as p:
            p.stdin.write(cert)
            decrypted_cert, stderr = p.communicate()
            log.info(stderr.decode("utf-8"))
            if p.returncode != 0:
                raise RuntimeError("Could not load encrypted certificate")
            return decrypted_cert
    else:
        log.warn("SSL certificate is plain text, consider gpg encrypting it.")
        return cert

def get_pubkey_as_der(pem):
    """Extract the public key from a private pem formatted cert formatted as DER"""
    with subprocess.Popen(
            ['openssl', 'rsa', '-pubout', '-outform', 'DER'],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE) as p:
        p.stdin.write(pem)
        der, err = p.communicate()
        return der

def get_pubkey_digest(der):
    """Get the sha256 digest of the pubkey in der format"""
    with subprocess.Popen(
            ['openssl', 'dgst', '-sha256', '-binary'],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE) as p:
        p.stdin.write(der)
        digest, err = p.communicate()
        return digest

def create_ssl_ctx(pem):
    """Create temporary file to store plain-text cert and create ssl
    context from it.

    This doesn't seem really secure since it requires using the
    filesystem, but I can't see another way as the low-level openssl
    api requires a file and will not accept a string or file like
    object.

    """
    with tempfile.NamedTemporaryFile('wb') as temp_cert:
        temp_cert.write(pem)
        temp_cert.flush()
        del pem
        ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_ctx.load_cert_chain(temp_cert.name)
        return ssl_ctx

def get_pinned_cert_hash(pem):
    """Get the public key from the private one and output the base64
    encoded sha256 hash that curl needs to pin a certificate"""
    der = get_pubkey_as_der(pem)
    digest = get_pubkey_digest(der)
    return str(base64.b64encode(digest), "utf-8")

def create_self_signed_cert():
    """Create a new self-signed cerificate"""
    with tempfile.NamedTemporaryFile() as temp_cert:
        with subprocess.Popen(
                ['openssl', 'req', '-new', '-x509', '-days', '365',
                '-nodes', '-out', temp_cert.name, '-keyout', temp_cert.name],
                stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE) as p:
            p.stdin.write(b"\n"*10)
        return open(temp_cert.name,'br').read()

def get_ssl_context_from_settings(settings):    
    if settings['ssl'] is not False:
        if settings['ssl'] is None:
            # Create self-signed certificate for one use:
            log.warn("No SSL certificate provided, creating a new self-signed certificate for this session")
            cert = create_self_signed_cert()
            # Always pin the certificate if we are using self-signed cert:
            settings['pin'] = True
        else:
            # Use pre-generated certificate file:
            with open(settings['ssl'], 'br') as cert_file:
                cert = decrypt_cert_if_necessary(cert_file.read())
        ssl_ctx = create_ssl_ctx(cert)
        settings['ssl_hash'] = get_pinned_cert_hash(cert)
        log.info("SSL certificate loaded")
    else:
        ssl_ctx = None
    return ssl_ctx



import logging
import subprocess
import requests
import re
import tempfile
from io import BytesIO

log = logging.getLogger('curlbomb.gpg')

class GPGExceptionError(Exception):
    pass

def detect_encrypted_resource(resource):
    """Detect if resource is encrypted with GnuPG

    Determining factors:
     a) The resource is ascii-armored text
     b) The resource is binary data starting with 0x8502"""
    try:
        # Read file like
        line = resource.read(50)
        resource.seek(0)
    except AttributeError:
        # Read str like
        line = resource[:50]
    if line.startswith(b"-----BEGIN PGP"):
        return True
    elif line.startswith(b'\x85\x02'):
        return True
    else:
        return False
        
def decrypt_resource_if_necessary(resource):
    """Decrypt resource and return as decrypted BytesIO.

    If resource is not encrypted, return original resource"""
    if not detect_encrypted_resource(resource):
        return resource
    log.warn("Resource appears to be encrypted, handing off to gpg for decryption...")
    with subprocess.Popen(
            ['gpg','-d'], stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE) as p:
        p.stdin.write(resource.read())
        decrypted, stderr = p.communicate()
    if p.returncode != 0:
        raise RuntimeError("Resource failed decryption")
    decrypted_resource = BytesIO(decrypted)
    return decrypted_resource

def encrypt_resource_to_recipients(resource, recipients):
    """Encrypt resource to a list of GPG IDs"""
    cmd = ['gpg', '--batch', '-e', *['-r {}'.format(r) for r in recipients]]
    log.debug(cmd)
    with subprocess.Popen(cmd,
                          stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE) as p:
        p.stdin.write(resource.read())
        encrypted, stderr = p.communicate()
        if p.returncode != 0:
            log.error(stderr)
            raise RuntimeError("Resource failed to encrypt (public-key)")
        encrypted_resource = BytesIO(encrypted)
        return encrypted_resource
    
def encrypt_resource_symmetric(resource, passphrase):
    """Encrypt resource with passphrase"""
    with subprocess.Popen(['gpg', '--batch', '-c','--passphrase-fd','0'],
                          stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE) as p:
        p.stdin.write(bytes(passphrase, "utf-8"))
        p.stdin.write(b"\n")
        p.stdin.write(resource.read())
        encrypted, stderr = p.communicate()
        if p.returncode != 0:
            log.error(stderr)
            raise RuntimeError("Resource failed to encrypt (symmetric)")
        encrypted_resource = BytesIO(encrypted)
        return encrypted_resource
    
def gpg_key_info(identifier):
    with subprocess.Popen(
            ['gpg','-k',identifier],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE) as p:
        out, err = p.communicate()
        if p.returncode != 0:
            raise GPGExceptionError("Unknown key for id: {}".format(identifier))
        return out

def verify_resource(resource, signature, allowed_authors=[]):
    """Verify signature is valid for resource

    allowed_authors is a list of GPG IDs that the signature is allowed to be from.
    If len(allowed_authors) == 0, any valid signature works.
    """
    if re.match("http[s]?://", signature):
        # Download signature:
        log.info("Downloading signature file: {}".format(signature))
        r = requests.get(signature)
        sig = r.content
    else:
        with open(signature, 'rb') as f:
            sig = f.read()

    with tempfile.NamedTemporaryFile() as sig_file, \
         tempfile.NamedTemporaryFile() as resource_file:
        # Write signature and resource to temp files:
        sig_file.write(sig)
        sig_file.flush()
        resource_file.write(resource.read())
        resource_file.flush()
        resource.seek(0)

        log.debug("signature temp file: {}".format(sig_file.name))
        log.debug("resource temp file: {}".format(resource_file.name))

        with subprocess.Popen(
                ['gpg', '--verify', sig_file.name, resource_file.name],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE) as p:
            out, err = p.communicate()
            err = str(err, "utf-8")
            log.debug("gpg returned: {}".format(p.returncode))
            if p.returncode == 0:
                m = re.search("^Primary key fingerprint: (.*)", err, re.MULTILINE)
                if m:
                    fingerprint = m.groups()[0]
                    fingerprint_info = gpg_key_info(fingerprint)
                    for author in allowed_authors:
                        if gpg_key_info(author) == fingerprint_info:
                            log.warn("Matched good signature to allowed author"
                                     ": {}".format(author))
                            return True
                    if len(allowed_authors) == 0:
                        log.warn("Resource matched good signature to: {}".format(fingerprint))
                        return True
                else:
                    raise GPGExceptionError(
                        "GPG did not list a valid fingerprint even though it returned 0")
    return False



import glob
import os
import shlex
import subprocess
import logging

log = logging.getLogger('curlbomb.put')

def add_parser(subparsers):
    put_parser = subparsers.add_parser(
        'put', help='Copy local files or directories to the client')
    put_parser.add_argument('source', metavar="SOURCE", nargs=1,
                            help="Local path to copy (or put glob in quotes)")
    put_parser.add_argument('dest', metavar="DEST", nargs='?',
                            help="Remote directory to copy to")
    put_parser.add_argument('--exclude', metavar="PATTERN", action='append',
                            help="Exclude files matching PATTERN, "
                            "a glob(3)-style wildcard pattern", default=[])
    put_parser.set_defaults(prepare_command=prepare)


def prepare(args, settings, parser):
    path = glob.glob(args.source[0])[0]
    parent_path, path = os.path.split(os.path.abspath(path))
    exclude_args = " ".join(["--exclude='{}'".format(p) for p in args.exclude])
    cmd = shlex.split('tar czh {exclude} -C "{parent_path}" "{path}"'.format(
        parent_path=parent_path, path=path, exclude=exclude_args))
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    args.resource = settings['resource'] = p.stdout
    if args.dest:
        settings['shell_command'] = 'cd "{dest}" && tar xzvf'.format(dest=args.dest)
    else:
        settings['shell_command'] = 'tar xzvf'
